"""parse token list to recursive JsonNodes"""
from typing import List
from pyjsonedit.token_list import TokenList

class JsonNode:
    """ json node with begin end end children"""
    def __init__(self, obj_type, **kwargs):
        self.type = obj_type
        self.start = kwargs['start']
        self.end = kwargs['end']
        self.name = kwargs.get('name','')
        self.kids = kwargs.get('kids',[])
        if 'value' in kwargs:
            self.kids = [kwargs['value']]

    def append(self,obj):
        """add one child"""
        self.kids.append( obj )

    def __repr__(self):
        return self.to_string()

    def to_string(self,depth=0):
        """this object as string"""
        prefix=' '*(2*depth)
        if self.type=='value':
            return f'JsonNode::{self.type}[{self.start}..{self.end}] = {self.kids[0]}'

        return f'JsonNode::{self.type}[{self.start}..{self.end}] ' +\
               f'\n{prefix}{{\n' +\
               ',\n'.join([ ' '*(2*(depth+1))+
                             str(i)+
                             ( '="'+child.name+'"' if child.name else '' )+
                             ": "+child.to_string(depth+1) for i,child in enumerate(self.kids)]) +\
               f'\n{prefix}}}'

    def __eq__(self, obj):
        return isinstance(obj, JsonNode) and\
                obj.start == self.start and\
                obj.end == self.end and\
                obj.type == self.type and\
                obj.kids == self.kids and\
                obj.name == self.name


def eat_value(tok: TokenList) -> JsonNode:
    """
    convert value tokens to object
    V -> JsonNode
    """
    begin = tok.expect_pop('v', 'not a value')
    return JsonNode('value',
                    start= begin[1],
                    end  =(begin[1] + len(begin[2])),
                    value= begin[2])

def eat_string(tok: TokenList) -> JsonNode:
    """
    convert stirng tokens to object
    V -> JsonNode
    """
    begin = tok.expect_pop('S', 'not a string')
    return JsonNode('value',
                    start = begin[1],
                    end   =(begin[1] + len(begin[2]) + 2),
                    value = begin[2])

def eat_child(tok: TokenList) -> JsonNode:
    """
    (S|V|A|D)
    """
    ret = None
    if   tok.next_is('v'):
        ret = eat_value(tok)
    elif tok.next_is('['):
        ret = eat_array(tok)
    elif tok.next_is('{'):
        ret = eat_dict(tok)
    elif tok.next_is('S'):
        ret = eat_string(tok)
    return ret

def eat_dict(tok:TokenList) -> JsonNode:
    """
    convert dict tokens to object,
    basicalyy array with named items
    '{' 'S':(V|A|D)* '}'
    """
    begin = tok.expect_pop('{', 'not object')
    ret = JsonNode('dict',
                   start=begin[1],
                   end=False)

    while tok.peek():

        if tok.next_is('S'):

            name = tok.pop()
            if len(name) != 3 or not name[2]:
                tok.raise_token_error('string token is missing value')
            tok.expect_pop(':', 'missing ":"')
            child = eat_child(tok)
            if not child:
                tok.raise_token_error('missing child')
            child.name = name[2]
            ret.append(child)

            if tok.next_is(','):
                tok.pop()
                continue

        end = tok.expect_pop('}', 'object error, unexpectd token: {}'.format( str(tok.peek()) ))
        ret.end = end[1]+1
        return ret

    tok.raise_token_error('object not closed')


def eat_array(tok:TokenList) -> JsonNode:
    """
    convert array tokens to object
    '[' ( S|V|A|D )* ']'
    """
    begin = tok.expect_pop('[', 'not array')

    ret = JsonNode('array',
                   start=begin[1],
                   end=False)
    while tok.peek():

        child = eat_child(tok)
        if child:
            ret.append(child)
            if tok.next_is(','):
                tok.pop()
                continue

        end = tok.expect_pop(']', 'array error unexpectd token: {}'.format( str(tok.peek()) ))
        ret.end = end[1]+1
        return ret

    tok.raise_token_error('array error, object not closed')

def parse(tokens:List) -> JsonNode:
    """
    Conver tokens into object (dict)
    """
    tok = TokenList(tokens)
    return eat_dict(tok)



"""module exports"""
from .tokenizer import tokenize
from .main import string_to_tokens,string_to_tree,string_match_mark
from .matcher import MatchException,match,print_matched
from .tree import JsonNode, parse



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyjsonedit']

package_data = \
{'': ['*']}

install_requires = \
['coverage==5.5', 'pylint==2.8.2', 'pytest==6.2.4']

entry_points = \
{'console_scripts': ['pyjsonedit = pyjsonedit.cli:run']}

setup_kwargs = {
    'name': 'pyjsonedit',
    'version': '0.0.16',
    'description': 'Edit parts of json strings & files while keeping the orginal, inconsistent formating',
    'long_description': '# pyJsonEdit\n\n[![PyPi version](https://badge.fury.io/py/pyjsonedit.svg)](https://pypi.org/project/pyjsonedit/)\n[![license](https://img.shields.io/github/license/Naereen/StrapDown.js.svg)]()\n[![tests](https://github.com/UrbanskiDawid/pyJsonEditor/actions/workflows/tests.yaml/badge.svg)](https://github.com/UrbanskiDawid/pyJsonEditor/actions/workflows/tests.yaml)\n\n[![](https://forthebadge.com/images/badges/made-with-python.svg)]()\n[![](https://forthebadge.com/images/badges/powered-by-coffee.svg)]()\n[![](https://forthebadge.com/images/badges/uses-badges.svg)]()\n\n\nEdit parts of inconsistently formatted json.\n\nIt\'s just a bit slower that doint this by hand!\n\n# matcher\n\nNow you can easly select **nodes** in json tree\n\nsyntax:\n\nselector | action | node type\n---------|--------|-------\n  *| select **all** items in current node| -\n [n] | select **n-th** item of curent node| array\n {n} | select **n-th** item of curent node| object\n key | select node chilld **by name**| object\n"key"| select node chilld **by name**| object\n \\>  | mark current node as seleced |-\n\n\nexample 1: \n\n```\nkey > [0]\n```\n\nthis pattern will match one element by:\n\n1. selecting "key" element in root node (assuring that is an object)\n2. select first element in it (assumintg its and array) \n\nexample 2: \n\n```\nname > *\n```\n\nthis pattern will match multiple elements by:\n\n1. selecting "name" element in root node (assuring that is an object)\n2. select all element in it \n\n## how to install\n\n```bash\npip install --upgrade pyjsonedit\n```\n\n## python module\n\n```python\n$ import pyjsonedit\n```\n## comand line\n\n```sh\n $ pyjsonedit --help\n```\n\n```bash\nUsage: pyjsonedit [OPTIONS] PATTERN JSON\n\n  cli method for masking matching parts of json\n\nOptions:\n  --symbol TEXT\n  --color        enable color output\n  --help         Show this message and exit.\n```\n\n## example: mask multiple nodes\n> $ pyjsonedit **"quiz > * > q1 >*"** DOC/example.json\n\n```\n{\n    "quiz": {\n        "sport": {\n            "q1": {\n                "question": XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX,\n                "options": XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX,\n                "answer": XXXXXXXXXXXXXXX\n            }\n        },\n        "maths": {\n            "q1": {\n                "question": XXXXXXXXXXX,\n                "options": XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX,\n                "answer": XXXX\n            },\n            "q2": {\n                "question": "12 - 8 = ?",\n                "options": [\n                    "1",\n                    "2",\n                    "3",\n                    "4"\n                ],\n                "answer": "4"\n            }\n        }\n    }\n}\n```\n\n## example: mask selected nodes\n\n```python\n$ import pyjsonedit\n$ pyjsonedit.string_match_mark("{\'pass\':123}","pass")\n{\'pass\':XXX}\n```\n\n[![string_match_mark](https://github.com/UrbanskiDawid/pyJsonEditor/raw/master/DOC/mask_pass.gif)]()\n',
    'author': 'Dawid Urbanski',
    'author_email': 'kontakt@dawidurbanski.pl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/UrbanskiDawid/pyJsonEditor',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""iterable list of tokens"""

class TokenError(Exception):
    """ parser exception"""

class TokenList:
    """helper class to iterate tokens"""
    def __init__(self, tok):
        self.tok = tok.copy()
        self.pos = 0

    def raise_token_error(self,comment:str):
        """ raport exceptions """
        raise TokenError(f'TokenError at postion:{self.pos} {comment}')

    def pop(self):
        """remove first token"""
        self.pos += 1
        return self.tok[self.pos-1]

    def peek(self):
        """see first token"""
        return self.tok[self.pos] if self.pos<len(self.tok) else None

    def next_is(self, tok_type):
        """check type on firts token"""
        return self.peek()[0] == tok_type if self.pos<len(self.tok) else False

    def expect_pop(self, tok_type, comment):
        """pop next item throw if unexpected token type"""
        if not self.peek() or not self.peek()[0] == tok_type:
            self.raise_token_error(comment)
        return self.pop()



#!/usr/bin/python3
"""main file to see execution"""

import click
from pyjsonedit import main

@click.argument('pattern')
@click.argument('json')
@click.option('--symbol', default='X', help='')
@click.option('--color', default=True, is_flag=True, help='enable color output')
def cli_match_mark(pattern, json, symbol, color):
    """cli method for masking matching parts of json"""
    main.cli_match_mark(pattern, json, symbol, color)

def run():
    """this method is used by package installer"""
    click.command()(cli_match_mark)()
